package ir.opt;

import ir.MyModule;
import ir.instruction.BrInstr;
import ir.value.BasicBlock;
import ir.value.Function;
import util.MyNode;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;

public class Analyze {
    // 计算紧前和后继基本块关系
    public static void calcuPrecSuccBB(MyModule module) {
        module.getFunctions().forEach(
                (functionMyNode -> calcuPrecSuccBBInFunc(functionMyNode.getValue()))
        );
    }

    private static void addEdge(BasicBlock prec, BasicBlock succ) {
        prec.getSuccBB().add(succ);
        succ.getPrecBB().add(prec);
    }
    private static void calcuPrecSuccBBInFunc(Function function) {
        function.getList().forEach(
                (bbNode) -> {
                    var bb = bbNode.getValue();
                    bb.clearPrecSuccInfo();
                    var instr = bb.getList().getLast().getValue(); // 最后一条指令是br或者ret
                    if (instr instanceof BrInstr) {
                        var brInstr = (BrInstr) instr;
                        if (brInstr.getBasicBlock1() != null) {
                            addEdge(bb, brInstr.getBasicBlock1());
                        }
                        if (brInstr.getBasicBlock2() != null) {
                            addEdge(bb, brInstr.getBasicBlock2());
                        }
                    }
                }
        );
    }

    // 计算支配信息
    // I'm told that this algorithm is on Engineering a Compiler P479.
    // So, here it is.
    public static void calcuDomination(MyModule module) {
        module.getFunctions().forEach(
                functionMyNode -> {
                    var function = functionMyNode.getValue();
                    calcuDominationForFunctionOpt(function);
                    calcuDominanceFrontierForFunction(function);
                }
        );

    }
    private static void calcuDominationForFunctionOpt(Function function) {
        var dom = new ArrayList<BitSet>();
        var index = new HashMap<BasicBlock, Integer>();
        var rindex = new ArrayList<BasicBlock>();
        var t = new BitSet();
        t.set(0);
        dom.add(t); // Dom(0) = {0}
        for (int i = 1; i < function.getList().getSize(); i++) {
            t = new BitSet();
            t.set(0,function.getList().getSize()-1);
            dom.add(t);
        } // forall n != 0, Dom(n) = N
        int cnt = 0;
        for (MyNode<BasicBlock> basicBlockMyNode : function.getList()) {
            var bb = basicBlockMyNode.getValue();
            bb.clearDominationInfo();
            index.put(bb,cnt);
            rindex.add(bb);
            cnt++;
        }
        boolean changed = true;
        while (changed) {
            changed = false;
            for (var bbNode :
                    function.getList()) {
                var bb = bbNode.getValue();
                var temp = new BitSet();
                for (var predbb :
                        bb.getPrecBB()) {
                    var i = index.get(predbb);
                    temp.and(dom.get(i));
                } //前驱dom的交
                var ndomIndex = index.get(bb);
                temp.set(ndomIndex); //并自己
                if (temp.equals(dom.get(ndomIndex))) {
                    dom.set(ndomIndex, temp);
                    changed = true;
                }
            }
        }

        // 把bitset的信息更新到BasicBlock
        for (int i = 0, rindexSize = rindex.size(); i < rindexSize; i++) {
            BasicBlock bb = rindex.get(i);
            var domer = dom.get(i).nextSetBit(0);
            while (domer != -1) {
                bb.getDominator().add(rindex.get(i));
                domer = dom.get(i).nextSetBit(domer + 1);
            }
        }
        function.getList().forEach(basicBlockMyNode -> {
            var bb = basicBlockMyNode.getValue();
            bb.getDominator().forEach(dombb -> dombb.getDominating().add(bb));
        }); // 计算被支配(dominating)信息

        // 计算immediate dominator
        for (int i = 0, domSize = dom.size(); i < domSize; i++) {
            BitSet domI = dom.get(i);
            var possibleIdom = domI.nextSetBit(0);

            while (possibleIdom != -1) {
                var isIdom = true;
                var otherNode = domI.nextSetBit(0);
                while (otherNode != -1) { //判断是不是其他必经节点的必经节点
                    if (otherNode != possibleIdom && dom.get(otherNode).get(possibleIdom)) {
                        isIdom = false;
                        break;
                    }
                    otherNode = domI.nextSetBit(otherNode + 1);
                }
                if (isIdom) {
                    rindex.get(i).setIdominator(rindex.get(possibleIdom));
                    break;
                }
                possibleIdom = domI.nextSetBit(possibleIdom + 1);
            }
        }

    }
    // 计算支配边界，需要先算出idom。
    private static void calcuDominanceFrontierForFunction(Function function) {
        for (var bbNode :
                function.getList()) {
            var bb = bbNode.getValue();
            for (var succ :
                    bb.getSuccBB()) {
                //bb 可能为null因为入口节点没有直接支配节点。
                while (bb != null && (!bb.getDominating().contains(succ) || bb == succ)) {
                    bb.getDf().add(succ);
                    bb = bb.getIdominator();
                }
            }
        }
    }
    // 这个算法效率略低，用bitset重新实现。
    private static void calcuDominationForFunction(Function function) {

        function.getList().forEach(
                (bbNode) -> {
                    var bb = bbNode.getValue();
                    bb.clearDominationInfo();
                    bb.getDominator().addAll(function.getList().toList());
                }
        ); // forall i != 0, Dom(i) = N
        var t = function.getList().getFirst().getValue().getDominator();
        t.clear();
        t.add(function.getList().getFirst().getValue()); // Dom(0) = {0}
        boolean[] changed = {true};
        while (changed[0]) {
            changed[0] = false;
            function.getList().forEach(
                    basicBlockMyNode -> {
                        var bb = basicBlockMyNode.getValue();
                        //var temp = new ArrayList<>(function.getList().toList());
                        var oldSize = bb.getDominator().size();
                        // 交集
                        for (var precBB :
                                bb.getPrecBB()) {
                            bb.getDominator().removeIf(e -> !precBB.getDominator().contains(e));
                        }
                        //temp.add(bb); // 并上自己
                        bb.getDominator().add(bb);
                        if (bb.getDominator().size() < oldSize) {
                            changed[0] = true;
                        }
                    }
            );
        }
        function.getList().forEach(basicBlockMyNode -> {
            var bb = basicBlockMyNode.getValue();
            bb.getDominator().forEach(dombb -> dombb.getDominating().add(bb));
        }); // 计算被支配信息
    }

}
