package parser;

import ir.type.*;
import org.antlr.v4.runtime.tree.*;
import treeir.*;
import util.Mylogger;

import java.util.*;
import java.util.logging.Logger;

/**
 * 实现了处理ANTLR产生的语法树的Vistor类
 */
public class MySysYVisitor extends AbstractParseTreeVisitor<Void> implements SysYVisitor<Void> {

    private static final MySysYVisitor instance = new MySysYVisitor();
    //private Logger logger = Mylogger.getLogger(MySysYVisitor.class);
    private Root root = new Root();
    private FuncDef funcDef = null;

    private Deque<Sequence> dom = new ArrayDeque<>(); // 嵌套的块级作用域，是栈，找局部变量从里往外找。
    private HashMap<String, TreeValue> globalConstants = new HashMap<>();
    private HashMap<String, SysYType> global = new HashMap<>();
    private HashMap<String, SysYType> functions = new HashMap<>();
    private SysYType defContextType = null; // 当前处理的声明的类型 比如 int a,b，那么parse <int>的时候先要把类型存下来

    private Deque<Integer> arrInitContextType = null; // e.g. int a[2][3] = {{}, {4}}; 需要用类型信息来推断应该如何初始化。
    private ArrayList<SysYType> funcFParamType = new ArrayList<>(); // 函数形参类型
    private Node current = root; // 当前节点
    public Root getRoot() {
        root.getFunctions().putAll(functions);
        root.getGlobals().putAll(global);
        return root;
    }
    /**
     * 获取当前可见的编译期可求值（const）变量
     *
     * @return Map
     */
    private Map<String, TreeValue> getConstants() {
        // 后放入哈希表的值会覆盖先放入的，实现作用域覆盖（好像不能覆盖？）
        var ret = new HashMap<String, TreeValue>();
        ret.putAll(globalConstants);
        if (funcDef != null) {
            ret.putAll(funcDef.getConstants());
        }
        for (var d :
                dom) {
            ret.putAll(d.getConstants());
        }
        return ret;
    }

    private void putConstant(String k, TreeValue v) {
        if (funcDef == null) {
            globalConstants.put(k,v);
        } else {
            if (dom.size() > 0) {
                dom.peek().getConstants().put(k,v);
            } else {
                funcDef.getConstants().put(k,v);
            }

        }
    }

    private void putDecl(String k, SysYType type) {
        if (funcDef == null) {
            global.put(k, type);
        } else {
            if (dom.size() > 0) {
                dom.peek().getLocals().put(k,type);
            } else {
                funcDef.getLocals().put(k,type);
            }
        }
    }

    @Override
    public Void visitCompUnit(SysYParser.CompUnitContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitDecl(SysYParser.DeclContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitConstDecl(SysYParser.ConstDeclContext ctx) {
        if (ctx.bType().INT() != null) {
            defContextType = IntSysYType.construct();
        } else if (ctx.bType().FLOAT() != null) {
            defContextType = FloatSysYType.construct();
        } else if (ctx.bType().VOID() != null) {
            //logger.severe("var decl using `void`");
            throw new Error("syntax error: var decl using `void`");
        }
        //defContextType = ConstSysYType.construct(defContextType);
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitConstDef(SysYParser.ConstDefContext ctx) {
        String ident = ctx.Ident().getText();
        SysYType compoundType = defContextType;
        ArrayList<Integer> dims = new ArrayList<>(); // 数组定义维度
        if (ctx.constExp() != null) {
            for (var exp :
                    ctx.constExp()) {
                visit(exp);
                var dimObj = current.getChildren().get(current.getChildren().size() - 1).evaluate(getConstants());
                current.getChildren().remove(current.getChildren().size() - 1);
                assert dimObj.discrimination == 0 || dimObj.discrimination == 1;
                if (dimObj.discrimination == 0) {
                    dims.add(dimObj.intVal);
                } else {
                    dims.add((int) dimObj.floatVal);
                }
            }
            for (int i = dims.size() - 1; i >= 0; i--) {
                compoundType = ArraySysYType.construct(compoundType, dims.get(i));
            }
        }
        var innerType = compoundType;
        compoundType = ConstSysYType.construct(compoundType);
        putDecl(ident, compoundType);

        if (ctx.constInitVal() != null) {
            if (innerType instanceof ArraySysYType) {
                var t = ((ArraySysYType)
                        ((ConstSysYType) compoundType).getParam())
                        .getDims();
                arrInitContextType = new ArrayDeque<>();
                for (int i = t.size() - 1; i >= 0; i--) {
                    arrInitContextType.push(t.get(i));
                }

                var ass = new Assign(new ArrayList<>(), current);
                current.addChildLast(ass);
                ass.addChildLast(new Ident(ass, ident));
                var arr = new MyArray(new ArrayList<>(), current);
                arr.setType(((ConstSysYType) compoundType).getParam());
                ass.addChildLast(arr);

                current = arr;
                visit(ctx.constInitVal());
                var val = arr.evaluate(getConstants());
                putConstant(ident, val);
                current = ass.getParent();

                arrInitContextType = null;
            } else {
                var ass = new Assign(new ArrayList<>(), current);
                current.addChildLast(ass);
                ass.addChildLast(new Ident(ass, ident));
                current = ass;
                visit(ctx.constInitVal());
                convetType();
                var val = ass.getChildren().get(1).evaluate(getConstants());
                putConstant(ident, val);
                current = ass.getParent();
            }
        }
        return null;
    }

    @Override
    public Void visitConstInitVal(SysYParser.ConstInitValContext ctx) {
        if (ctx.constExp() != null) {
            visit(ctx.constExp());
            convetType();
        } else {
            int startOffset = current.getChildren().size();
            //拆一层
            var dim = arrInitContextType.pop();

            for (var iv :
                    ctx.constInitVal()) {
                visit(iv);
            }

            //装回去
            arrInitContextType.push(dim);
            var elemOffset = 1;
            for (var i :
                    arrInitContextType) {
                elemOffset *= i;
            }
            // 不完整初始化列表补0
            int n = elemOffset - (current.getChildren().size() - startOffset);
            for (int i = 0; i < n; i++) {
                if (defContextType instanceof IntSysYType) {
                    current.addChildLast(new NumberInt(0, current));
                } else {
                    current.addChildLast(new NumberFloat(0, current));
                }
            }
        }
        return null;
    }

    private void convetType() {
        if (current.getLastChild() instanceof NumberFloat && defContextType instanceof IntSysYType) {
            var node = (NumberFloat)current.popLastChild();
            current.addChildLast(node.castInt());
        }
        if (current.getLastChild() instanceof NumberInt && defContextType instanceof FloatSysYType) {
            var node = (NumberInt)current.popLastChild();
            current.addChildLast(node.castFloat());
        }
    }

    @Override
    public Void visitVarDecl(SysYParser.VarDeclContext ctx) {
        if (ctx.bType().INT() != null) {
            defContextType = IntSysYType.construct();
        } else if (ctx.bType().FLOAT() != null) {
            defContextType = FloatSysYType.construct();
        } else if (ctx.bType().VOID() != null) {
            //logger.severe("var decl using `void`");
            throw new Error("syntax error: var decl using `void`");
        }
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitVarDef(SysYParser.VarDefContext ctx) {
        String ident = ctx.Ident().getText();
        SysYType compoundType = defContextType;
        ArrayList<Integer> dims = new ArrayList<>(); // 数组定义维度
        if (ctx.constExp() != null) {
            for (var exp :
                    ctx.constExp()) {
                visit(exp);
                var dimObj = current.getChildren().get(current.getChildren().size() - 1).evaluate(getConstants());
                current.getChildren().remove(current.getChildren().size() - 1);
                assert dimObj.discrimination == 0 || dimObj.discrimination == 1;
                if (dimObj.discrimination == 0) {
                    dims.add(dimObj.intVal);
                } else {
                    dims.add((int) dimObj.floatVal);
                }
            }
            for (int i = dims.size() - 1; i >= 0; i--) {
                compoundType = ArraySysYType.construct(compoundType, dims.get(i));
            }
        }
        putDecl(ident, compoundType);
        if (ctx.initVal() != null) {
            if (compoundType instanceof ArraySysYType) {
                var t = ((ArraySysYType)
                        compoundType)
                        .getDims();
                arrInitContextType = new ArrayDeque<>();
                for (int i = t.size()-1; i >= 0; i--) {
                    arrInitContextType.push(t.get(i));
                }

                var ass = new Assign(new ArrayList<>(), current);
                current.addChildLast(ass);
                ass.addChildLast(new Ident(ass, ident));
                var arr = new MyArray(new ArrayList<>(), current);
                arr.setType(compoundType);
                ass.addChildLast(arr);

                current = arr;
                visit(ctx.initVal());
                var val = arr.evaluate(getConstants());
                putConstant(ident, val);
                current = ass.getParent();

                arrInitContextType = null;
            } else {
                var ass = new Assign(new ArrayList<>(), current);
                current.addChildLast(ass);
                ass.addChildLast(new Ident(ass, ident));
                current = ass;
                visit(ctx.initVal());
                // 复制代码没看清楚，不应该尝试对非常量求值。
                //var val = ass.getChildren().get(1).evaluate(getConstants());
                //putConstant(ident, val);
                current = ass.getParent();
            }

        }
        return null;
    }

    @Override
    public Void visitInitVal(SysYParser.InitValContext ctx) {
        if (ctx.exp() != null) {
            visit(ctx.exp());
        } else {
            //拆一层
            var dim = arrInitContextType.pop();
            int startOffset = current.getChildren().size();
            for (var iv :
                    ctx.initVal()) {
                visit(iv);
            }

            //装回去
            arrInitContextType.push(dim);
            var elemOffset = 1;
            for (var i :
                    arrInitContextType) {
                elemOffset *= i;
            }
            // 不完整初始化列表补0
            int n = elemOffset - (current.getChildren().size() - startOffset);
            for (int i = 0; i < n; i++) {
                if (defContextType instanceof IntSysYType) {
                    current.addChildLast(new NumberInt(0, current));
                } else {
                    current.addChildLast(new NumberFloat(0, current));
                }
            }
        }
        return null;
    }

    @Override
    public Void visitFuncDef(SysYParser.FuncDefContext ctx) {
        var funcIdent = ctx.Ident().getText();
        var func = new FuncDef(current, funcIdent);
        funcDef = func;
        current.addChildLast(func);
        current = func;
        if (ctx.funcFParams() != null) {
            visit(ctx.funcFParams());
        }
        visit(ctx.funcType());
        var funcRetType = defContextType;


        visit(ctx.block());
        var type = new FunctionSysYType(funcFParamType, funcRetType);
        func.setType(type);
        functions.put(funcIdent, type);
        current = current.getParent();

        funcDef = null;
        return null;
    }

    @Override
    public Void visitFuncType(SysYParser.FuncTypeContext ctx) {
        if (ctx.INT() != null) {
            defContextType = IntSysYType.construct();
        } else if (ctx.FLOAT() != null) {
            defContextType = FloatSysYType.construct();
        } else {
            defContextType = VoidSysYType.construct();
        }
        return null;
    }

    @Override
    public Void visitFuncFParams(SysYParser.FuncFParamsContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitFuncFParam(SysYParser.FuncFParamContext ctx) {
        assert current instanceof FuncDef;
        var func = (FuncDef) current;
        func.getParameter().add(ctx.Ident().getText());
        SysYType baseType = null;
        if (ctx.bType().INT() != null) {
            baseType = IntSysYType.construct();
        } else if (ctx.bType().FLOAT() != null) {
            baseType = FloatSysYType.construct();
        } else if (ctx.bType().VOID() != null) {
            //logger.severe("var decl using `void`");
            throw new Error("syntax error: var decl using `void`");
        }
        SysYType compoundType = baseType;
        ArrayList<Integer> dims = new ArrayList<>(); // 数组定义维度
        if (ctx.funcFArrParam() != null) {

            for (var exp :
                    ctx.funcFArrParam().exp()) {
                visit(exp);
                var dimObj = current.getChildren().get(current.getChildren().size() - 1).evaluate(globalConstants);
                current.getChildren().remove(current.getChildren().size() - 1);
                assert dimObj.discrimination == 0 || dimObj.discrimination == 1;
                if (dimObj.discrimination == 0) {
                    dims.add(dimObj.intVal);
                } else {
                    dims.add((int) dimObj.floatVal);
                }
            }
            for (int i = dims.size() - 1; i >= 0; i--) {
                compoundType = ArraySysYType.construct(compoundType, dims.get(i));
            }
            compoundType = ArraySysYType.construct(compoundType, -1); //形参第一个维数为空 eg. int para[][2]
        }
        func.getLocals().put(ctx.Ident().getText(), compoundType);
        return null;
    }

    @Override
    public Void visitFuncFArrParam(SysYParser.FuncFArrParamContext ctx) {
        return null;
    }

    @Override
    public Void visitBlock(SysYParser.BlockContext ctx) {
        var seq = new Sequence(new ArrayList<>(), current);
        dom.push(seq);
        current.addChildLast(seq);
        current = seq;
        visitChildren(ctx);
        current = seq.getParent();
        dom.pop();
        return null;
    }

    @Override
    public Void visitBlockItem(SysYParser.BlockItemContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitAssign(SysYParser.AssignContext ctx) {
        var assign = new Assign(new ArrayList<>(), current);
        current.addChildLast(assign);
        current = assign;
        visit(ctx.lVal());
        visit(ctx.exp());
        current = current.getParent();
        return null;
    }

    @Override
    public Void visitExpr(SysYParser.ExprContext ctx) {
        if (ctx.exp() == null) {
            var t = new NumberInt(0, current);
            current.addChildLast(t);
        }
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitBlockStmt(SysYParser.BlockStmtContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitIf(SysYParser.IfContext ctx) {
        var cond = new Condition(new ArrayList<>(), current);
        current.addChildLast(cond);
        current = cond;
        visit(ctx.cond());
        visit(ctx.stmt().get(0)); // if branch
        if (ctx.stmt().size() > 1) {
            visit(ctx.stmt().get(1)); // else branch
        }
        current = current.getParent();
        return null;
    }

    @Override
    public Void visitWhile(SysYParser.WhileContext ctx) {
        var loop = new Loop(new ArrayList<>(), current);
        current.addChildLast(loop);
        current = loop;
        visit(ctx.cond());
        visit(ctx.stmt());
        current = current.getParent();
        return null;
    }

    @Override
    public Void visitBreak(SysYParser.BreakContext ctx) {
        //assert current instanceof Loop; 可能在循环的更深作用域里
        current.addChildLast(new Break(current));
        return null;
    }

    @Override
    public Void visitContinue(SysYParser.ContinueContext ctx) {
        //assert current instanceof Loop; 可能在循环的更深作用域里
        current.addChildLast(new Continue(current));
        return null;
    }

    @Override
    public Void visitReturn(SysYParser.ReturnContext ctx) {
        //assert current instanceof FuncDef; 还可能是Seq
        var ret = new Return(new ArrayList<>(), current);
        current.addChildLast(ret);
        current = ret;
        if (ctx.exp() != null) {
            visit(ctx.exp());
        } // 可能 return;
        current = current.getParent();
        return null;
    }

    @Override
    public Void visitExp(SysYParser.ExpContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitCond(SysYParser.CondContext ctx) {
        visitChildren(ctx);
        return null;
    }

    /**
     * 因为Index的计算方式，树是反着构造的。
     *
     * @param ctx the parse tree
     * @return Void
     */
    @Override
    public Void visitLVal(SysYParser.LValContext ctx) {
        var ident = ctx.Ident().getText();
        if (ctx.exp() == null) {
            current.addChildLast(new Ident(current, ident));
        } else {
            var index = new Index(new ArrayList<>(), null);
            //current.addChildLast(index);
            var trueCurrent = current;
            current = index;
            index.addChildLast(new Ident(index, ident));
            for (var exp :
                    ctx.exp()) {
                visit(exp);
                var t = new Index(new ArrayList<>(), null);
                current.setParent(t);
                t.addChildLast(current);
                current = t;
            }
            var top = current.getChildren().get(0); // 多套了一层
            top.setParent(trueCurrent);
            trueCurrent.addChildLast(top);
            current = trueCurrent;
        }
        return null;
    }

    @Override
    public Void visitPrimaryExp(SysYParser.PrimaryExpContext ctx) {
        if (ctx.exp() != null) {
            visit(ctx.exp());
        } else if (ctx.String() != null) {
            var t = ctx.String().getText();
            t = t.substring(1,t.length()-1); // 移出首尾引号
            current.addChildLast(new MyString(t,current));
        } else if (ctx.lVal() != null) {
            visit(ctx.lVal());
        } else if (ctx.number() != null) {
            visit(ctx.number());
        }
        return null;
    }

    @Override
    public Void visitNumber(SysYParser.NumberContext ctx) {
        if (ctx.floatConst() != null) {
            current.addChildLast(new NumberFloat(Float.parseFloat(ctx.floatConst().getText()), current));
        } else if (ctx.intConst() != null) {
            var t = 0;
            if (ctx.intConst().Dec() != null) {
                t = Integer.parseInt(ctx.intConst().Dec().getText());
            } else if (ctx.intConst().Hex() != null) {
                t = Integer.parseInt(ctx.intConst().Hex().getText().substring(2), 16); // should remove 0x
            } else if (ctx.intConst().Oct() != null) {
                var ts = ctx.intConst().Oct().getText();
                if (ts.length() != 1) {
                    t = Integer.parseInt(ts.substring(1), 8);
                }
                // should remove 0, unless IT IS 0
            }
            current.addChildLast(new NumberInt(t, current));
        }
        return null;
    }

    @Override
    public Void visitUnaryExp(SysYParser.UnaryExpContext ctx) {
        if (ctx.primaryExp() != null) {
            visit(ctx.primaryExp());
        } else if (ctx.unaryExp() != null) {
            var op = new Operation(new ArrayList<>(), new ArrayList<>(), current, null);
            switch (ctx.unaryOp().getText()) {
                case "+" -> {
                    visit(ctx.unaryExp());
                }
                case "-" -> {
                    op.appendOp(Operation.Operator.Neg);
                    current.addChildLast(op);
                    current = op;
                    visit(ctx.unaryExp());
                    var inner = current.getChildren().get(0);
                    if (inner instanceof Operation && ((Operation) inner).getOperators().get(0) == Operation.Operator.Neg) {
                        var innermost = inner.getChildren().get(0);
                        innermost.setParent(current.getParent());
                        current.getParent().popLastChild();
                        current.getParent().addChildLast(innermost);
                    }
                    current = current.getParent();
                }
                case "!" -> {
                    op.appendOp(Operation.Operator.Not);
                    current.addChildLast(op);
                    current = op;
                    visit(ctx.unaryExp());
                    var inner = current.getChildren().get(0);
                    if (inner instanceof Operation && ((Operation) inner).getOperators().get(0) == Operation.Operator.Not) {
                        var innermost = inner.getChildren().get(0);
                        innermost.setParent(current.getParent());
                        current.getParent().popLastChild();
                        current.getParent().addChildLast(innermost);
                    }
                    current = current.getParent();
                }
            }
        } else if (ctx.Ident() != null) {
            var funCall = new FunCall(new ArrayList<>(), current, ctx.Ident().getText());
            current.addChildLast(funCall);
            current = funCall;
            if (ctx.funcRParams() != null) { //可能空参调用
                visit(ctx.funcRParams());
            }

            current = current.getParent();
        }
        return null;
    }

    @Override
    public Void visitFuncRParams(SysYParser.FuncRParamsContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Void visitMulExp(SysYParser.MulExpContext ctx) {
        if (ctx.unaryExp().size() == 1) {
            visit(ctx.unaryExp(0));
        } else {
            var op = new Operation(new ArrayList<>(), new ArrayList<>(), current, null);
            current.addChildLast(op);
            current = op;
            visit(ctx.unaryExp(0));
            for (int i = 1; i < ctx.unaryExp().size(); i++) {
                visit(ctx.unaryExp(i));
                switch (ctx.mulOp().get(i-1).getText()) {
                    case "*" -> {
                        op.appendOp(Operation.Operator.Mul);
                    }
                    case "/" -> {
                        op.appendOp(Operation.Operator.Div);
                    }
                    case "%" -> {
                        op.appendOp(Operation.Operator.Mod);
                    }
                }
            }
            current = current.getParent();
        }
        return null;
    }

    @Override
    public Void visitAddExp(SysYParser.AddExpContext ctx) {
        if (ctx.mulExp().size() == 1) {
            visit(ctx.mulExp(0));
        } else {
            var op = new Operation(new ArrayList<>(), new ArrayList<>(), current, null);
            current.addChildLast(op);
            current = op;
            visit(ctx.mulExp(0));
            for (int i = 1; i < ctx.mulExp().size(); i++) {
                visit(ctx.mulExp(i));
                switch (ctx.addOp(i-1).getText()) {
                    case "+" -> {
                        op.appendOp(Operation.Operator.Add);
                    }
                    case "-" -> {
                        op.appendOp(Operation.Operator.Sub);
                    }
                }
            }
            current = current.getParent();

        }
        return null;
    }

    @Override
    public Void visitRelExp(SysYParser.RelExpContext ctx) {
        if (ctx.addExp().size() == 1) {
            visit(ctx.addExp(0));
        } else {
            var op = new Operation(new ArrayList<>(), new ArrayList<>(), current, null);
            current.addChildLast(op);
            current = op;
            visit(ctx.addExp(0));
            for (int i = 1; i < ctx.addExp().size(); i++) {
                visit(ctx.addExp(i));
                switch (ctx.relOp(i-1).getText()) {
                    case "<" -> {
                        op.appendOp(Operation.Operator.Lt);
                    }
                    case ">" -> {
                        op.appendOp(Operation.Operator.Gt);
                    }
                    case "<=" -> {
                        op.appendOp(Operation.Operator.Le);
                    }
                    case ">=" -> {
                        op.appendOp(Operation.Operator.Ge);
                    }
                }
            }
            current = current.getParent();
        }
        return null;
    }

    @Override
    public Void visitEqExp(SysYParser.EqExpContext ctx) {
        if (ctx.relExp().size() == 1) {
            visit(ctx.relExp(0));
        } else {
            var op = new Operation(new ArrayList<>(), new ArrayList<>(), current, null);
            current.addChildLast(op);
            current = op;
            visit(ctx.relExp(0));
            for (int i = 1; i < ctx.relExp().size(); i++) {
                visit(ctx.relExp(i));
                switch (ctx.eqOp(i-1).getText()) {
                    case "==" -> {
                        op.appendOp(Operation.Operator.Eq);
                    }
                    case "!=" -> {
                        op.appendOp(Operation.Operator.Ne);
                    }
                }
            }
            current = current.getParent();
        }
        return null;
    }

    @Override
    public Void visitLAndExp(SysYParser.LAndExpContext ctx) {
        if (ctx.eqExp().size() == 1) {
            visit(ctx.eqExp(0));
        } else {
            var op = new Operation(new ArrayList<>(), new ArrayList<>(), current, null);
            current.addChildLast(op);
            current = op;
            visit(ctx.eqExp(0));
            for (int i = 1; i < ctx.eqExp().size(); i++) {
                visit(ctx.eqExp(i));
                op.appendOp(Operation.Operator.And);
            }
            current = current.getParent();
        }
        return null;
    }

    @Override
    public Void visitLOrExp(SysYParser.LOrExpContext ctx) {
        if (ctx.lAndExp().size() == 1) {
            visit(ctx.lAndExp(0));
        } else {
            var op = new Operation(new ArrayList<>(), new ArrayList<>(), current, null);
            current.addChildLast(op);
            current = op;
            visit(ctx.lAndExp(0));
            for (int i = 1; i < ctx.lAndExp().size(); i++) {
                visit(ctx.lAndExp(i));
                op.appendOp(Operation.Operator.And);
            }
            current = current.getParent();
        }
        return null;
    }

    @Override
    public Void visitConstExp(SysYParser.ConstExpContext ctx) {
        visitChildren(ctx);
        // 编译期计算常量表达式
        var val = current.getChildren().get(current.getChildren().size()-1).evaluate(getConstants());
        current.getChildren().remove(current.getChildren().size()-1);
        if (val.discrimination == 0) {
            current.getChildren().add(new NumberInt(val.intVal, current));
        } else {
            current.getChildren().add(new NumberFloat(val.floatVal, current));
        }
        return null;
    }

    @Override
    public Void visitMulOp(SysYParser.MulOpContext ctx) {
        return null;
    }

    @Override
    public Void visitAddOp(SysYParser.AddOpContext ctx) {
        return null;
    }

    @Override
    public Void visitRelOp(SysYParser.RelOpContext ctx) {
        return null;
    }

    @Override
    public Void visitEqOp(SysYParser.EqOpContext ctx) {
        return null;
    }

    @Override
    public Void visitBType(SysYParser.BTypeContext ctx) {
        return null;
    }

    @Override
    public Void visitUnaryOp(SysYParser.UnaryOpContext ctx) {
        return null;
    }

    @Override
    public Void visitIntConst(SysYParser.IntConstContext ctx) {
        return null;
    }

    @Override
    public Void visitFloatConst(SysYParser.FloatConstContext ctx) {
        return null;
    }

    private static class VarInfo {
        public int slot;
        public SysYType type;
    }

}
